#!/usr/bin/env python

# pyADuCprog (c) 2005 Thomas Langewouters, ALFA quality software
# Licensed under GPL v2

from heximage import Hex_Image
import serial
import time 
import struct

class Aducprog:
	
	class ChecksumError ( IOError ):
		"Checksum Mismatch!"
		pass
	class AcknowledgeError ( IOError ):
		"Host sent a NACK byte"
		pass

	def __init__(self):
		self.port = ''
		self.log = ''

	def checksum(self, data):
		s = sum(data)
		checksum = (0x100 - s) & 0xFF
		return checksum

	def Open_Port(self, port = ''):
		if isinstance(self.port, serial.Serial):
			self.port.close()
		self.Log('Opening %s ...' % port)
		self.port = serial.Serial(port, timeout=10)

	def DownloadHex(self, filename):
		self.CheckConnected()	
		Image = Hex_Image(filename)
		#Codegeheugen wissen (nodig)
		self.EraseProgramMemory()
		
		at = 0	# beginadres
		packet = []	# [...data...]
		packets = {}	# { adres1 : [data1,..,data7],
				#   adres2 : [data1,..,data7], ... }

		# Geheugenbereik hexfile aflopen en in paketten verdelen
		
		for Address in range(0,0xF7FF): #len(Image.Data)
			# huidige byte gewoon toevoegen aan paket
                        byte = Image.Get( Address, Default = 0xFF )
			packet.append(byte)
			#print "adr: %s" % Address 
			if len(packet) == 1:
				# adres van huidige byte onthouden
				FirstAddress = Address 
				#print "newadr: %s" % FirstAddress
			if len(packet) == 8:
				# paket zit vol, toevoegen aan hoop,
				# met als key het adres van de 1e byte.
				packets[FirstAddress] = packet
				packet = []
		#het onafgewerkte paket toevoegen
		packets[FirstAddress] = packet
		for startaddress,packet in sorted(packets.iteritems()):
			#print "s: index %s \ns: data %s" %(startaddress, packet)
			# startadres omzetten naar drie bytes (int):
			startaddressbytes = [startaddress / (2**16), \
			(startaddress % (2**16)) / (2**8), ((startaddress % (2**16)) % (2**8))]
			#print "start address %s, addressbytes %s" %(startaddress,startaddressbytes) 
			packet = startaddressbytes + packet
			self.WriteProgramBlock(packet)	

		self.port.close()	

	def CheckConnected(self):
		if not isinstance(self.port, serial.Serial):
			self.Open_Port()

		#self.port.write(struct.pack('B',0x06))
		#time.sleep(0.2)
		self.port.flushInput()
		self.port.flushOutput()
		time.sleep(0.1)
		self.port.write(struct.pack('BBBB',0x21,0x5A,0x00,0xA6))
		time.sleep(0.1)
		
		if not self.options.testing:
			result = self.port.read(25)
		else:
			self.Log('!!!! FAKING received response')
			result = struct.pack('25B',65, 68, 73, 32, 56, 51, 50, 32, 32, 32, 86, 50, 48, 32, 13, 10, 120, 136, 204, 1, 5, 163, 1, 0, 176)
	
		# result => array of bytes
		# failsafe as er niets ontvangen is...
		if len(result) > 0:
			result = list(struct.unpack('%iB' % len(result), result))
			
		self.Log('Received packet: ' + str(result))
		
		# kijk of het paket met ADI begint..
		if len(result) < 3 or result[:3] != [0x41,0x44,0x49]:
			raise Exception, "Identification string didn't start with ADI"

		# Checksum verificatie.
		mychecksum = 0x100 - (sum(result[:-1]) & 0xFF)
		spchecksum = result[-1:].pop()	# zonder de pop krijg ik [int] ipv int
		self.Log('verifying mychecksum %#x with %#x...' % (mychecksum, spchecksum))

		if mychecksum != spchecksum: 
			raise self.ChecksumError, 'my: %#x sp: %#x' % (mychecksum, spchecksum)

		#TODO: dit in een lus gieten en een delay inlassen zodat
		#TODO: de gebruiker de controller alsnog kan resetten.
		#TODO: iets nuttig doen met het informatiepaket
		self.Log("identification packet received succesfully!")
		self.Log('Sending ACK.')
		self.port.write(struct.pack('B',0x06))
		return 1

	def SendPacket(self,data):
		# zou nooit mogen gebeuren
		if len(data) > 25:
			raise Exception, "Can't send packets longer than 25"

		frame = [0x07,0x0E]
		frame += data
		frame.append(self.checksum(data))


		self.Log("SendFrame:" + len(frame) * ' %#x' % tuple(frame)) 

		# to string for sending
		frame = struct.pack('%iB'%len(frame),*frame)
		# FIXME DEBUG
		if self.options.testing:
			#self.Log("Emulating sending and receiving")
			response = 0x06	
		else:
			self.port.write(frame)
			response = self.port.read(1)
			if len(response) == 0:
					raise Exception, 'no reply'
			response = struct.unpack('B',response)
		
		if response == 0x06:
			pass
		elif response == 0x07: 
			#Nack
			raise Exception, 'NACK on frame %s' % ascframe
		else:
			raise Exception, 'Invalid response %s on frame %s'\
			% (struct.unpack('B',response),ascframe)
			
	def EraseProgramMemory(self):
		"Send an Erase Program Memory Only Packet"
		self.Log('Sending \"Erase Program Memory\"!')
		self.SendPacket([0x43])	# C

	def EraseProgramAndData(self):
		"Send an Erase Program and Data memory Packet"
		self.Log('Sending \"Erase Program and data\"!')
		self.SendPacket([0x41])	# A

	def WriteProgramBlock(self,data):
		"Commit a program block to program memory"
		# FIXME
		self.Log( "Write CODE:  @ %#x %#x %#x -> "  % tuple(data[:3]) + \
		(len(data) -3) * " %#x" % tuple(data[3:]) )
		data.insert(0,0x57)	# W
		self.SendPacket(data)

	def WriteDataBlock(self,data):
		"Commit a data block to data memory"
		self.Log("Write DATA: %x %x %x -> " + (len(data) -3)*" %x" % \
		data )
		self.SendPacket(data.insert(0,0x45)) # E
	
	def Log( self, String ):
		"debug log"
		#self.log = self.log + String + '\n'
		print time.strftime('%H:%I:%M:%S ') + String

if __name__ == '__main__':
	"parse command Line and obey the commands"
	from optparse import OptionParser
	parser = OptionParser()
	
	parser.add_option("-p", "--port",  dest="port", \
	default='/dev/tts/USB0',  help="Use this serial port",\
	metavar="FILE")
	
	parser.add_option("-q", "--quiet", dest="verbose", \
	action="store_false",default=True, \
	help="don't print status messages to stdout")
	
	parser.add_option("-s", "--sayhi", dest="sayhi", \
	action="store_true",default=False, \
	help="Check if device is connected")

	parser.add_option("-t", "--testing", dest="testing", \
	action="store_true",default=False, \
	help="Don't communicate with device - for testing purposes")

	parser.add_option("-b", "--burn", dest="burn", \
	action="store",default='', type='string', \
	help="Check if device is connected")

	(options, args) = parser.parse_args()
	programmer = Aducprog()
	programmer.options = options
	programmer.Open_Port(options.port)
		
	if options.sayhi:
		programmer.CheckConnected()
	if options.burn <> '':
		programmer.DownloadHex(options.burn)
